from __future__ import unicode_literals
from __future__ import absolute_import
import logging
from django.core.exceptions import ImproperlyConfigured
from django import forms
from django.conf import settings
from . import utils, registry

logger = logging.getLogger(__name__)


class SiteConfigBase(object):

    application_short_name = "default_application"
    application_verbose_name = "Default Application"

    def get_backend(self):
        backend = getattr(settings, 'SITECONFIG_BACKEND_DEFAULT',
            'site_config.backends.model_backend.DatabaseBackend')
        return backend

    def get_default_configs(self):
        """
        Returns a configuration dictionary of configuration variables and their defaults.
        The dictionary keys set the configuration variables.
        The dictionary values are another nested dictionary.
          This nested dictionary must contain 3 keys:
            default = the default value that the key will take
            field = a django Field instance used to validate the value
            help (optional) = a help text entry that describes the key

        The configuration values should be upper-case by convention.
        """
        return {'EXAMPLE_A': {'default': "Test A default",
                             'field': 'django.forms.CharField',
                             'help': 'Test A help text.'},
                "EXAMPLE_B": {'default': 1,
                             'field': 'django.forms.IntegerField',
                             'help': 'Test B help text.'}}

    def __init__(self, website=None):
        self.website = website
        for k, v in self.get_default_configs().items():
            if 'default' not in v:
                raise ImproperlyConfigured("Config value dict %s must have a 'default' key." % (v))
            elif 'field' not in v:
                raise ImproperlyConfigured(
                    "Config value dict %s must have a 'field' key, which is a valid django field." % (v))
        super(SiteConfigBase, self).__setattr__('_backend',
            utils.import_module_attr(self.get_backend())())

    def __getattr__(self, name):
        if name in list(self.get_default_configs().keys()):
            return self.get_config(name)['value']
        else:
            raise AttributeError("%r object has no attribute %r" %
                         (self.__class__, name))

    def get_config(self, key,):
        """
        This method gets the configuration dictionary from get_default_configs()
        and looks up the provided key from that dictionary, which returns the nested
        configuration dictionary for that key.

        That nested dictionary is then passed to the storage backend to be
        updated with a new 'value' key.  That value key is either the result
        of the backend lookup, or the value of the 'default' key.

        Example Return Value:
                {'default':"Test A default", 'field':forms.CharField,
                           'help':'Test A help text.', 'value':"looked-up value A"}
        """
        result = self._backend.get(key, self.get_default_configs(),
                                   self.application_short_name, self.website, )
        return result

    def get_configs(self):
        """
        This method gets the configuration dictionary from get_default_configs()
        and passes it to the storage backend.

        The storage backend returns a similar dictionary with the value key added to the
        nested dictionaries.  These value keys are the result of the configuration lookup
        or the default specified by the default nested dictionary keys.
        Example Return Value:
                {"EXAMPLE_A":{'default':"Test A default",
                              'field':forms.CharField,
                              'help':'Test A help text.',
                              'value':"looked-up value A"},
                 "EXAMPLE_B":{'default':1,
                              'field':forms.IntegerField,
                              'help':'Test B help text.',
                              'value':"looked-up value B"}}
        """
        return self._backend.mget(self.get_default_configs(), self.application_short_name, self.website)

    def set_config(self, key, value,):
        self._backend.set(key, value, self.get_default_configs(), self.application_short_name, self.website)

    def website_application_status(self):
        return self._backend.website_application_status(self.application_short_name, self.website)

    def get_curtain_message(self):
        return self._backend.get_curtain_message(self.application_short_name, self.website)

    def __dir__(self):
        return list(self.get_default_configs().keys())


# registry.config_registry.register(SiteConfigBase)




WEBAPP_ACTIVE_STATE_DISABLED = "disabled"
WEBAPP_ACTIVE_STATE_CURTAINED = "curtained"
WEBAPP_ACTIVE_STATE_ENABLED = "enabled"

WEBAPP_ACTIVE_STATES = (
  (WEBAPP_ACTIVE_STATE_DISABLED, "Disabled"),
  (WEBAPP_ACTIVE_STATE_CURTAINED, "Curtained"),
  (WEBAPP_ACTIVE_STATE_ENABLED, "Enabled"),
)



from __future__ import unicode_literals
from __future__ import absolute_import
import logging
from site_config import utils, choices, registry
from django.conf import settings
from .. import ConfigBackend

logger = logging.getLogger(__name__)


class SettingsBackend(ConfigBackend):

    def get(self, key, config_dict, application_short_name, website_short_name=None):
        config_meta = config_dict[key]
        config_meta.update({'value': getattr(settings, key, config_meta['default'])})
        return config_meta

    def website_application_status(self, application_short_name, website_short_name=None):
        return getattr(settings, 'SITECONFIG_SITEAPP_STATUS', choices.WEBAPP_ACTIVE_STATE_ENABLED)

    def get_curtain_message(self, application_short_name, website_short_name=None):
        return getattr(settings, 'SITECONFIG_CURTAIN_MESSAGE',
                       "This site is undergoing scheduled maintenance."
                       "Thank you for your patience.")



from __future__ import unicode_literals
from __future__ import absolute_import
import logging
from jsonfield import JSONField
import django
from django.utils.encoding import python_2_unicode_compatible
from django.db import models
from site_config import utils, choices

logger = logging.getLogger(__name__)


@python_2_unicode_compatible
class Website(models.Model):

    name = models.CharField(
        max_length=64,
        help_text="Enter a descriptive name for this website.")
    short_name = models.SlugField(
        max_length=64, unique=True,
        help_text="This must be a unique name using only "
                  "letter, numbers, hyphens, and underscores.")
    active = models.BooleanField(
        default=False,
        help_text="Enable or disable the entire website.")
    description = models.TextField(blank=True, null=True)

    class Meta:
        app_label = 'site_config'
        verbose_name = "Website"
        verbose_name_plural = "Websites"
        ordering = ['short_name']

    def __str__(self):
        return "%s (%s)" % (self.name, self.short_name)


@python_2_unicode_compatible
class Application(models.Model):

    short_name = models.SlugField(
        max_length=64, unique=True,
        help_text="This must be a unique name using only "
                  "letter, numbers, hyphens, and underscores.")
    active = models.BooleanField(
        default=False,
        help_text="Enable or disable the entire application.")
    description = models.TextField(blank=True, null=True)

    class Meta:
        app_label = 'site_config'
        verbose_name = "Application"
        verbose_name_plural = "Applications"
        ordering = ['short_name']

    def __str__(self):
        return "%s" % (self.short_name)


#class WebsiteApplicationQuerySet(models.query.QuerySet):
#
#    def active(self):
#        return self.filter(active=True)
#
#    def website_applications(self, website_short_name, application_short_name):
#        return self.filter(
#                        application__short_name=application_short_name,
#                        website__short_name=website_short_name, )

# Why was this done? Are we creating a custom Manager object and overriding its
# "__getattr__()" method just so apps can call "objects.website_applications()"
# instead of the normal "objects.filter( ... )"? (see this pkg's "__init__.py").
# for an example of where this was done). Overriding "__getattr__()" can lead to
# trouble, at least in Dj. 1.11.
#
# If I've overlooked something and things don't work right anymore, I'm sorry.
# Hopefully I can fix it better later. - NTT

#class WebsiteApplicationManager(models.Manager):
#
#    def get_queryset(self):
#        return WebsiteApplicationQuerySet(self.model, using=self._db)
#
#    if django.VERSION < (1, 6):
#        get_query_set = get_queryset
#
#
# Specifically, stuff breaks when the `get()` method of `class DatabaseBackend`
# (see this pkg's "__init__.py") calls its `mget()` method, which then has:
#
#     site_app_list = models.WebsiteApplication.objects.website_applications(...
#
# Django then throws "AttributeError: 'WebsiteApplicationManager' object has no
# attribute 'website_applications'"
#
# The reason it thinks the attribute `website_applications` doesn't exist when
# it is defined above is because `__getattr__()` is overridden below.
#
##    def __getattr__(self, attr, *args):
##        try:
##            return getattr(self.__class__, attr, *args)
##        except AttributeError:
##            get_queryset = (
##                self.get_query_set
##                if hasattr(self, 'get_query_set')
##                else self.get_queryset)
##            return getattr(get_queryset(), attr, *args)


@python_2_unicode_compatible
class WebsiteApplication(models.Model):

    website = models.ForeignKey('site_config.Website', on_delete=models.CASCADE)
    application = models.ForeignKey('site_config.Application', on_delete=models.CASCADE)

    active = models.CharField(
        max_length=20,
        default=choices.WEBAPP_ACTIVE_STATE_DISABLED,
        choices=choices.WEBAPP_ACTIVE_STATES,
        help_text="Activates, curtains or deactivates this website application "
        "combination. In order for this to be active, both the corresponding "
        "website and application must also be active.  Curtained sites can "
        "only be viewed by superusers.")
    description = models.TextField(blank=True, null=True)
    curtain_message = models.TextField(
        blank=True, null=True,
        default="This site is undergoing scheduled maintenance."
                "Thank you for your patience.",
        help_text="Specify a message that should be displayed "
                  "when the site is in the curtained state.")
    options = JSONField(blank=True, null=True)

# See comments above for why I deactivated this custom manager - NTT.
#    objects = WebsiteApplicationManager()

    def get_config_options(self, default_config_dict):
        return utils.update_config_dict(default_config_dict, self.options)

    def set_config_options(self, config_dict, save=True):
        self.options = config_dict
        if save:
            self.save()

    def active_status(self):
        return_value = choices.WEBAPP_ACTIVE_STATE_DISABLED
        if self.website.active and self.application.active and \
           self.active in [choices.WEBAPP_ACTIVE_STATE_ENABLED,
                           choices.WEBAPP_ACTIVE_STATE_CURTAINED]:
            return_value = self.active
        return return_value

    def get_curtain_message(self):
        return self.curtain_message

    class Meta:
        app_label = 'site_config'
        verbose_name = "Website Application"
        verbose_name_plural = "Websites Applications"
        unique_together = (('website', 'application'),)

    def __str__(self):
        return "%s (%s)" % (self.website, self.application)



from __future__ import unicode_literals
from __future__ import absolute_import
import importlib
import logging
from django import forms
from site_config import registry, utils
from . import models

logger = logging.getLogger(__name__)


def website_application_formfactory(instance=None):
    config_class = None
    meta_options = {
        "model": models.WebsiteApplication,
        'exclude': ['options', ],
    }
    properties = {"Meta": type(b'Meta', (), meta_options)}
    # for deserialization, we need to know which fields are config fields
    config_fields = []
    # only add config options for existing objects
    if instance:
        # lookup the configuration class for this object,
        # based on the application short_name
        config_lookup = registry.config_registry.get_config_class(
            instance.application.short_name
        )
        if config_lookup:
            config_class = config_lookup[1](
                website=instance.website.short_name
            )
            config_fields = []
            for config_name, lookup_dict in config_class.get_configs().items():
                config_fields.append(config_name)
                field_class = utils.import_module_attr(lookup_dict['field'])
                kwargs = dict(
                    label=config_name,
                    help_text="%s Default: %s" % (lookup_dict.get('help', ''),
                                                  lookup_dict['default']),
                    initial=lookup_dict['value'],
                    required=False
                )
                if hasattr(field_class, 'choices') and \
                        'choices' in lookup_dict:
                    kwargs.update(
                        {'choices': lookup_dict.get('choices', None)}
                    )
                if hasattr(field_class, "queryset") and \
                        'queryset' in lookup_dict:
                    func = import_helper(lookup_dict.get('queryset'))
                    kwargs.update(
                        {'queryset': func()}
                    )
                properties.update({
                    config_name: field_class(**kwargs),
                })
            properties.update(
                {
                    'reset_options': forms.BooleanField(
                        label="Reset to Defaults",
                        required=False, initial=False
                    ),
                }
            )

    def clean_form(self):
        cleaned_data = super(self.__class__, self).clean()
        if instance and config_class:
            if cleaned_data.get('reset_options'):
                cleaned_data['options'] = utils.config_dict_value_from_default(
                    config_class.get_configs())
            else:
                cleaned_configs = {}
                for k, v in cleaned_data.items():
                    if k in config_fields:
                        cleaned_configs.update(
                            {k: {'value': cleaned_data.pop(k)}}
                        )
                cleaned_data['options'] = utils.update_config_dict(
                    config_class.get_configs(), cleaned_configs)
        return cleaned_data
    properties.update(
        {"config_fields": sorted(config_fields), 'clean': clean_form})
    form = type(
        b'WebsiteApplicationAdminForm', (forms.ModelForm, ), properties
    )
    return form


def import_helper(name):
    class_data = name.split(".")
    module_path = ".".join(class_data[:-1])
    class_str = class_data[-1]
    module = importlib.import_module(module_path)
    return getattr(module, class_str)



from __future__ import unicode_literals

from .utils import import_module_attr


class ConfigRegistry():

    def __init__(self):
        self._registry = {}

    def register(self, siteconfig):
        self._registry[siteconfig.application_short_name] = (
            siteconfig.application_verbose_name, siteconfig)

    def get_config_list(self):
        return [(k, v[0]) for k, v in self._registry.items()]

    def get_config_class(self, application_short_name):
        return self._registry.get(application_short_name, None)


config_registry = ConfigRegistry()



from __future__ import unicode_literals
from __future__ import absolute_import
import os
from copy import deepcopy
from django.db.models.query import QuerySet
try:
    from importlib import import_module
except ImportError:
    from django.utils.importlib import import_module
from django.template.loader import select_template


def import_module_attr(path):
    package, module = path.rsplit('.', 1)
    return getattr(import_module(package), module)


def id_for_item(item):
    if hasattr(item, 'id'):
        return item.id
    else:
        return item


def update_config_dict(default_config_dict, updated_config_dict):
    """
    Takes a config_dict and updates the values in the nested meta
    dicts with the values provided in the updated_config_dict.

    INPUT
     default_config_dict = see the output of
          site_config.get_default_configs()
     updated_config_dict = {'TEST_A': {"value": "1234"},
                           'TEST_B': {"value": 1},
                           }
    OUTPUT
    i.e.
    {"TEST_A": {"default": ...,
                "field": ...,
                "help": ...,
                "value":"1234",
                }
     "TEST_B": {"default": ...,
                "field": ...,
                "help": ...,
                "value":1,
                }
    }
    """
    config_dict = deepcopy(default_config_dict)
    if type(updated_config_dict) is dict:
        # update the value keys to be the contents of updated_config_dict
        for config_name, nested_config_dict in config_dict.items():
            if config_name in updated_config_dict and \
                    'value' in updated_config_dict[config_name]:
                if 'queryset' in config_dict[config_name]:
                    """
                    If it is a list of queryset, make sure we serialize as
                    list of integers to prevent JSON serialization problems.

                    If it is an object with an id, make sure we serialize as
                    the id of the object.
                    """
                    val = updated_config_dict[config_name]['value']
                    if isinstance(val, list) or isinstance(val, QuerySet):
                        val = [id_for_item(v) for v in val]
                    else:
                        val = id_for_item(val)
                    config_dict[config_name].update(
                        {'value': val}
                    )
                else:
                    config_dict[config_name].update(
                        {'value': updated_config_dict[config_name]['value']}
                    )
    return config_dict


def config_dict_value_from_default(default_config_dict):
    """
    Takes a config_dict and updates the "value" attribute of
    all of the nested "meta" dicts to the value of the corresponding
    "default" attribute.

    INPUT
     default_config_dict = see the output of
          site_config.get_default_configs()
    """

    new_config_dict = deepcopy(default_config_dict)
    # set a default 'value' in each nested config dict
    for config_name, x in new_config_dict.items():
        new_config_dict[config_name].update(
            {'value': new_config_dict[config_name]['default']})
    return new_config_dict


def website_override_template(template_name, website):
    """
    Tries to look for a template on the template path named
    [website]/[template_name], then falls back to looking
    for a template at [template_name].  If neither exist,
    this raises a TempateDoesNotExist error.
    """
    website_template_name = os.path.join(website, template_name)
    template_obj = select_template((website_template_name, template_name,))
    # Prior to Django 1.8, the above returns a template.
    # With Django 1.8, we need to unwrap the template from the above.
    if hasattr(template_obj, 'template'):
        template_obj = template_obj.template
    return template_obj


class WebsiteOverrideTemplateViewMixin(object):
    """
    Mix this class into a CBV where you want to
    provide a means for the website to override the
    template. The self.website attribute must be
    set in order for this mixin to override a template.

    Tries to look for a template on the template path named
    [self.website]/[template_name], then falls back to looking
    for a template at [template_name].  If neither exist,
    this raises a TempateDoesNotExist error.
    """
    def get_template_names(self):
        templates = super(
            WebsiteOverrideTemplateViewMixin, self
        ).get_template_names()
        website = getattr(self, 'website', None)
        if website is None:
            website = self.kwargs.get('website', None)
        if website:
            templates = [os.path.join(website, self.template_name)] + templates
        return templates



from __future__ import absolute_import
from django.test import TestCase
from django.test.utils import override_settings


import mock

from site_config import context_processors


class TestDecideBaseTemplate(TestCase):
    @override_settings(**{'SITECONFIG_BASE_TEMPLATE': "test.html"})
    def test_setting_override(self):
        """
        Verify default template used if setting not overridden
        """
        request = mock.Mock()
        request.resolver_match.kwargs.get.return_value = None
        request.path = '/'
        context = context_processors.decide_base_template(request)
        self.assertEqual(context['base_template'], "test.html")

    def test_setting_default(self):
        """
        Verify setting override changes default template
        """
        request = mock.Mock()
        request.resolver_match.kwargs.get.return_value = None
        request.path = '/'
        context = context_processors.decide_base_template(request)
        self.assertEqual(context['base_template'], "base_site.html")

    @mock.patch('site_config.context_processors.website_override_template')
    def test_website_pulled_from_kwargs(self, template_override_mock):
        """
        Verify website pulled from resolver_match kwargs passed to lookup
        """
        request = mock.Mock()
        request.resolver_match.kwargs.get.return_value = 'site-1'
        context_processors.decide_base_template(request)
        template_override_mock.assert_called_with(
            'base_site.html', 'site-1'
        )

    @mock.patch('site_config.context_processors.website_override_template')
    def test_website_not_set_if_not_match(self, template_override_mock):
        """
        If this would raise a 404, verify override not called
        """
        request = mock.Mock()
        request.resolver_match.kwargs.get.side_effect = Exception('something')
        request.path = '/'
        context_processors.decide_base_template(request)
        template_override_mock.assert_not_called()

    @mock.patch('site_config.context_processors.website_override_template')
    def test_template_lookup_result_returned(self, template_override_mock):
        """
        Verify result of website_override_template returned by function
        """
        mock_template = mock.Mock()
        mock_template.name = 'site-1/base_site.html'
        request = mock.Mock()
        request.resolver_match.kwargs.get.return_value = 'site-1'
        template_override_mock.return_value = mock_template
        context = context_processors.decide_base_template(request)
        self.assertEqual(
            context['base_template'], 'site-1/base_site.html'
        )

    @mock.patch('site_config.context_processors.website_override_template')
    def test_website_pulled_from_path_if_no_site_kwarg(self,
                                                       template_override_mock):
        """
        If website name isn't in kwargs, extract as first element of URL path
        """
        mock_template = mock.Mock()
        mock_template.name = 'site-2/base_site.html'
        template_override_mock.return_value = mock_template
        request = mock.Mock()
        request.path = '/site-2/something/kinda/long/'
        request.resolver_match.kwargs.get.side_effect = Exception('something')
        context = context_processors.decide_base_template(request)
        template_override_mock.assert_called_once_with(
            'base_site.html', 'site-2')
        self.assertEqual(
            context['base_template'], 'site-2/base_site.html'
        )



from __future__ import unicode_literals


class ConfigBackend(object):

    def get(self, key, config_dict, application_short_name, website_short_name=None):
        """
        Get the key from the backend store and return the value.
        Return None if not found.
        
        RETURNS:
        This returns the appropriate meta dictionary back (for the 
        provided key), with the 'value' key added.  
        
        i.e
           {'default':"1234", "field":"django.forms.CharField", 
            "value":"5432", "help":"this is the help text."}
        
        """
        raise NotImplementedError

    def mget(self, config_dict, application_short_name, website_short_name=None):
        """
        Get all the configuration values set for this application.
        
        RETURNS:
        This returns the config_dict with 'value' keys added to 
        each meta dictionary.
        
        i.e. 
          {
            "TEST_1":{
                'default':"1234", 
                "field":"django.forms.CharField", 
                "value":"5432", 
                "help":"this is the help text."},
            "TEST_2":{
                'default':"1234", 
                "field":"django.forms.CharField", 
                "value":"5432", 
                "help":"this is the help text."}
           }
        """
        raise NotImplementedError

    def set(self, config_name, value, config_dict, application_short_name, website_short_name=None):
        """
        Add the value to the backend store given the key.
        """
        raise NotImplementedError

    def mset(self, config_dict, application_short_name, website_short_name=None):
        """
        This saves all of the config values to the backend.
        """
        raise NotImplementedError

    def website_application_status(application_short_name, website_short_name=None):
        """
        This returns whether or not the provided website application is active,
        is disabled, or is curtained. 
        It returns one of the strings defined in choices.WEBAPP_ACTIVE_STATES
        """
        raise NotImplementedError

    def get_curtain_message(self, application_short_name, website_short_name=None):
        """
        Returns a string that is the message that an end user should see
        when the site is curtained. 
        """
        raise NotImplementedError


#!/usr/bin/env python
from __future__ import absolute_import
import os
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='site_config',
    version='1.1.0',
    description='Django configuration Utility to manage multiple "websites" in a project. ',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    author='Joe Jasinski',
    author_email='jjasinski@imagescape.com',
    install_requires=[
        'django >= 1.11',
        'jsonfield >= 0.9.22',
    ],
    url='https://github.com/ImaginaryLandscape/django_site_config',
    keywords='django site config configuration',
    extras_require={
        'testing': ["mock >= 1.0.1", "six >= 1.9.0"],
    },
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Framework :: Django",
        "Topic :: Utilities",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: BSD License",
    ],
)
